from __future__ import annotations

from typing import ClassVar, Literal, Tuple, overload

from .config import Config
from .ctx import Context, Result, Satisfier
from .node import ROOT, Node, NodeState
from .owner import Owner, export

# 分为两类方法：针对权限节点的操作，与针对权限状态的操作
# 权限状态: get, set
#   sget: root 获取执行者权限状态
#       1. 对于目标节点，若其父节点不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常或返回 None
#       2. 对于目标节点，若其自身不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常或返回 None
#   get: 执行者获取目标权限状态
#       1. 对于目标节点，若其父节点不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常或返回 None
#       2. 执行者: 对于目标节点，若其父节点的权限不包含 v+a，则抛出异常
#       3. 对于目标节点，若其自身不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常或返回 None
#       4. 执行者: 对于目标节点，若其自身的权限不包含 v，则抛出异常，否则返回该节点的状态
#   sset: root 设置权限状态
#       1. 对于目标节点，若其父节点不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       2. 对于目标节点，若其自身不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#   set: 执行者设置目标权限状态
#       1. 对于目标节点，若其父节点不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       2. 执行者: 对于目标节点，若其父节点的权限不包含 v+m+a，则抛出异常
#       3. 对于目标节点，若其自身不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       4. 执行者: 对于目标节点，若其自身的权限不包含 m，则不会修改该节点的状态


def _get(nodes: dict[Context, NodeState], context: Context, satisfier: Satisfier):
    return {ctx: state for ctx, state in nodes.items() if satisfier(context, ctx)}


class PermissionExecutor:
    def __init__(self, executor: Owner):
        self.executor = executor

    def _check_self(self, node: Node, expected: Tuple[NodeState, NodeState], context: Context, satisfier: Satisfier):
        _node = node.resolve()
        _executor_nodes = export(self.executor)
        while _node.parent != ROOT:
            node_parent = _node.parent
            if node_parent in _executor_nodes:
                states = _get(_executor_nodes[node_parent], context, satisfier)
                state = max(states.values())
                if state & expected[0] != expected[0]:
                    return False
            elif Config.DEFAULT_DIR & expected[0] != expected[0]:
                return False
            _node = node_parent
        if node in _executor_nodes:
            states = _get(_executor_nodes[node], context, satisfier)
            state = max(states.values())
            if state & expected[1] != expected[1]:
                return False
        elif Config.DEFAULT_DIR & expected[1] != expected[1]:
            return False
        return True

    @overload
    def get(
        self, target: Owner, node: Node | str, context: Context | None = None, satisfier: Satisfier | None = None
    ) -> Result[NodeState]: ...

    @overload
    def get(
        self,
        target: Owner,
        node: Node | str,
        context: Context | None = None,
        satisfier: Satisfier | None = None,
        missing_ok: Literal[True] = True,
    ) -> Result[NodeState] | None: ...

    def get(
        self,
        target: Owner,
        node: Node | str,
        context: Context | None = None,
        satisfier: Satisfier | None = None,
        missing_ok: bool = False,
    ):
        """获取节点状态

        Args:
            target (Owner): 目标对象
            node (Node | str): 目标节点
            context (Context, optional): 上下文. Defaults to None.
            satisfier (Satisfier, optional): 上下文筛选器. Defaults to None.
            missing_ok (bool, optional): 是否允许不存在. Defaults to False.
        """
        _node = node.resolve() if isinstance(node, Node) else Node(node).absolute()
        if not _node.exists():
            if missing_ok:
                return
            raise FileNotFoundError(f"node {_node} not exists")
        _ctx = context or Context.current()
        _satisfier = satisfier or Satisfier.all()
        if not self._check_self(_node, (NodeState("v-a"), NodeState("v--")), _ctx, _satisfier):
            raise PermissionError(f"permission denied for {self.executor} to access {_node}")
        _nodes = export(target)
        if _node not in _nodes:
            return Result({_ctx: Config.DEFAULT_DIR if _node.is_dir() else Config.DEFAULT_FILE}, _ctx)
        if not _ctx.data:
            return Result(_nodes[_node], _ctx)
        return Result(
            _get(_nodes[_node], _ctx, _satisfier)
            or {_ctx: Config.DEFAULT_DIR if _node.is_dir() else Config.DEFAULT_FILE},
            _ctx,
        )

    def set(
        self,
        target: Owner,
        node: Node | str,
        state: NodeState,
        context: Context | None = None,
        satisfier: Satisfier | None = None,
        missing_ok: bool = False,
        recursive: bool = False,
    ):
        """设置节点状态

        Args:
            target (Owner): 目标对象
            node (Node | str): 目标节点
            state (NodeState): 目标状态
            context (Context, optional): 上下文. Defaults to None.
            satisfier (Satisfier, optional): 满足器. Defaults to None.
            missing_ok (bool, optional): 是否允许不存在. Defaults to False.
            recursive (bool, optional): 是否递归. Defaults to False.
        """
        _node = node if isinstance(node, Node) else Node(node).absolute()
        if _node.stem == "*":
            _node = _node.parent
            recursive = True
        if not _node.exists():
            if missing_ok:
                return
            raise FileNotFoundError(f"node {_node} not exists")
        _ctx = context or Context.current()
        _satisfier = satisfier or Satisfier.all()
        if not self._check_self(_node, (NodeState("vma"), NodeState("-m-")), _ctx, _satisfier):
            raise PermissionError(f"permission denied for {self.executor} to modify {target}'s permission")
        target.nodes.setdefault(_node, {})[_ctx] = state
        if recursive and _node.is_dir():
            for node in _node.iterdir():
                self.root.set(
                    target, node, state, context, satisfier, missing_ok, recursive if node.is_dir() else False
                )

    root: ClassVar["RootPermissionExecutor"]


class RootPermissionExecutor(PermissionExecutor):
    def __init__(self):
        pass

    def _check_self(self, node: Node, expected: Tuple[NodeState, NodeState], context: Context, satisfier: Satisfier):
        return True


PermissionExecutor.root = RootPermissionExecutor()

# 权限节点：create, delete, modify, move, copy, list, require
#   create: 创建权限节点
#       1. 对于目标节点，若其父节点不存在，则依据参数 `parent: bool` 决定是否创建父节点，否则抛出异常
#       2. 对于目标节点，若其父节点前的父节点的权限不包含 v+a，则抛出异常
#       3. 对于目标节点，若其父节点的权限不包含 v+m+a，则抛出异常
#       4. 对于目标节点，若其自身已存在，则依据参数 `exist_ok: bool` 决定是否抛出异常
#   delete: 删除权限节点
#       1. 对于目标节点，若其父节点不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       2. 对于目标节点，若其父节点前的父节点的权限不包含 v+a，则抛出异常
#       3. 对于目标节点，若其父节点的权限不包含 v+m+a，则抛出异常
#       4. 对于目标节点，若其自身不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#   modify: 修改权限节点内容
#       1. 对于目标节点，若其父节点不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       2. 对于目标节点，若其父节点前的父节点的权限不包含 v+a，则抛出异常
#       3. 对于目标节点，若其父节点的权限不包含 v+m+a，则抛出异常
#       4. 对于目标节点，若其自身不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       5. 对于目标节点，若其自身的权限不包含 m，则抛出异常
#   move: 移动权限节点
#       1. 对于原始节点，若其父节点不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       2. 对于原始节点，若其父节点前的父节点的权限不包含 v+a，则抛出异常
#       3. 对于原始节点，若其父节点的权限不包含 v+m+a，则抛出异常
#       4. 对于原始节点，若其自身不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       5. 对于目标节点，若其父节点不存在，则依据参数 `parent: bool` 决定是否创建父节点，否则抛出异常
#       6. 对于目标节点，若其父节点前的父节点的权限不包含 v+a，则抛出异常
#       7. 对于目标节点，若其父节点的权限不包含 v+m+a，则抛出异常
#       8. 对于目标节点，若其自身已存在，则依据参数 `exist_ok: bool` 决定是否抛出异常
#   copy: 复制权限节点
#       1. 对于原始节点，若其父节点不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       2. 对于原始节点，若其父节点前的父节点的权限不包含 v+a，则抛出异常
#       3. 对于原始节点，若其父节点的权限不包含 v+m+a，则抛出异常
#       4. 对于原始节点，若其自身不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       5. 对于原始节点，若其自身的权限不包含 v，则抛出异常
#       6. 对于目标节点，若其父节点不存在，则依据参数 `parent: bool` 决定是否创建父节点，否则抛出异常
#       7. 对于目标节点，若其父节点前的父节点的权限不包含 v+a，则抛出异常
#       8. 对于目标节点，若其父节点的权限不包含 v+m+a，则抛出异常
#       9. 对于目标节点，若其自身已存在，则依据参数 `exist_ok: bool` 决定是否抛出异常
#   list: 列出权限节点
#       1. 对于目标节点，若其父节点不存在，则依据参数 `missing_ok: bool` 决定是否抛出异常
#       2. 对于目标节点，若其父节点前的父节点的权限不包含 v+a，则抛出异常
#       3. 对于目标节点，若其父节点的权限不包含 v+a，则抛出异常
#       4. 对于目标节点，若其自身的权限不包含 v，则抛出异常，否则返回该节点的子节点的状态
#   require: 判断权限节点是否存在
#       1. 对于目标节点，若其父节点不存在，则返回 False
#       2. 对于目标节点，若其父节点前的父节点的权限不包含 v+a，则返回 `Unknown`
#       3. 对于目标节点，若其父节点的权限不包含 v+a，则返回 `Unknown`
#       4. 对于目标节点，存在则返回 True，否则返回 False



from .ctx import Context as Context
from .ctx import context as context
from .function import PermissionExecutor as PermissionExecutor
from .monitor import AsyncMonitor as AsyncMonitor
from .monitor import SyncMonitor as SyncMonitor
from .node import ROOT as ROOT
from .node import Node as Node
from .node import NodeState as NodeState
from .owner import Group as Group
from .owner import User as User

PE = PermissionExecutor



from __future__ import annotations

from typing import Protocol, Sequence, Union

from .ctx import Context
from .node import Node, NodeState


class Group(Protocol):
    name: str
    priority: int
    inherits: list[Group]
    nodes: dict[Node, dict[Context, NodeState]]


class User(Protocol):
    name: str
    inherits: list[Group]
    nodes: dict[Node, dict[Context, NodeState]]


Owner = Union[Group, User]


def iter_inherits(inherits: Sequence[Owner]):
    for owner in inherits:
        if owner.inherits:
            yield from iter_inherits(owner.inherits)
        yield owner


def export(owner: Owner) -> dict[Node, dict[Context, NodeState]]:
    nodes = {}
    gps = []
    if owner.inherits:
        gps.extend(iter_inherits(owner.inherits))
    gps = list(set(gps))
    gps.sort(key=lambda x: x.priority, reverse=True)
    gps.append(owner)
    for gp in gps:
        for node, data in gp.nodes.items():
            if node not in nodes:
                nodes[node] = data.copy()
            else:
                nodes[node].update(data)
    return nodes



from pprint import pprint

from arclet.cithun.node import CHILD_MAP, INDEX_MAP, mkdir

FOOD = mkdir("food")
FRUIT = (FOOD / "fruit").mkdir()
VEGETABLE = (FOOD / "vegetable").mkdir()
APPLE = (FRUIT / "apple").touch()
BANANA = (FRUIT / "banana").touch()
SMALL_TOMATO = (VEGETABLE / "tomato" / "small tomato").mkdir(parents=True)
print(SMALL_TOMATO.dot())
"""
# NodeTree
/
└── food
    ├── fruit
    │   ├── apple
    │   └── banana
    └── vegetable
        └── tomato
            └── small tomato
"""

pprint(INDEX_MAP)
pprint(CHILD_MAP)



from functools import wraps
from typing import Callable, TypeVar
from typing_extensions import Concatenate, ParamSpec

from arclet.cithun import NodeState, User

T = TypeVar("T")
P = ParamSpec("P")

user = User("cithun")


def require(path: str, missing_ok: bool = False) -> Callable[[Callable[P, T]], Callable[Concatenate[User, P], T]]:
    def decorator(func: Callable[P, T]) -> Callable[Concatenate[User, P], T]:
        @wraps(func)
        def wrapper(usr: User, *args: P.args, **kwargs: P.kwargs) -> T:
            if usr.require(path, missing_ok=missing_ok):
                return func(*args, **kwargs)
            else:
                raise PermissionError(f"Permission denied for {usr.name} to access {path}")

        return wrapper

    return decorator


@require("/foo/bar/baz")
def alice():
    print("alice")


@require("/foo/bar/baz/qux", missing_ok=True)
def bob():
    print("bob")


@require("/foo/bar/baz/qux", missing_ok=False)
def caven():
    print("caven")


user.sadd("/foo/bar/baz", NodeState("vma"))
alice(user)  # alice
bob(user)  # bob as target node's parent is available

try:
    caven(user)
except PermissionError as e:
    # raise PermissionError as caven specified missing_ok=False, and baz/qux is not exist
    print(e)

user.smodify("/foo/bar/", NodeState("v--"))

try:
    alice(user)
except PermissionError as e:
    print(e)  # raise PermissionError as alice's target node is not available



from collections import UserDict
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any, Callable, Dict, Generic, List, TypeVar


class Context(UserDict):
    @staticmethod
    def current() -> "Context":
        return _ctx.get(Context())

    def __hash__(self):
        _res = []
        _size = 0
        for k, v in self.data.items():
            _res.append(k.__hash__() ^ v.__hash__())
            _size += len(k) + len(v)
        return hash(tuple(_res) + (_size,))

    def __repr__(self):
        return f"[{'|'.join(f'{k}={v}' for k, v in self.data.items())}]" if self.data else "[none]"

    @classmethod
    def from_string(cls, data: str):
        _data = {}
        if data != "[none]":
            for seg in data[1:-1].split("|"):
                segs = seg.split("=")
                _data[segs[0]] = segs[1]
        return cls(**_data)


_ctx = ContextVar("context")


@contextmanager
def context(**kwargs):
    token = _ctx.set(Context(**kwargs))
    try:
        yield
    finally:
        _ctx.reset(token)


class Satisfier:
    """上下文筛选器，用于判断当前上下文与目标上下文是否满足一定关系"""

    @staticmethod
    def all():
        return Satisfier(lambda self, other: self.data == other.data)

    @staticmethod
    def least():
        return Satisfier(lambda self, other: any((other.get(k) == v for k, v in self.data.items())))

    def __init__(self, func: Callable[[Context, Context], bool]):
        self.func = func

    def __call__(self, target: "Context", other: "Context") -> bool:
        if not target.data:
            return True
        if not other.data:
            return True
        return self.func(target, other)


T = TypeVar("T")


class Result(Generic[T]):
    def __init__(self, data: Dict[Context, T], origin: Context):
        self.data = data
        self.origin = origin

    @property
    def most(self) -> T:
        if self.origin in self.data:
            return self.data[self.origin]
        sortd = sorted(
            self.data.keys(), key=lambda x: sum((x.get(k) == v for k, v in self.origin.items())), reverse=True
        )
        return self.data[sortd[0]]

    def query(self, **kwargs: Callable[[Any], bool]) -> List[T]:
        return [v for k, v in self.data.items() for _k, _v in kwargs.items() if _k in k and _v(k[_k])]

    def __repr__(self):
        return f"Result({self.data})"



from pathlib import Path

from arclet.cithun import PE, Node, NodeState, context
from arclet.cithun.builtins.monitor import DefaultMonitor
from arclet.cithun.node import CHILD_MAP

monitor = DefaultMonitor(Path("monitor.json"))

baz = Node("/foo/bar/baz").mkdir(parents=True)
qux = (baz / "qux").touch()
cmd = Node("/command").mkdir()
test = (cmd / "test").mkdir()
sub = (test / "sub").touch()
test1 = (cmd / "test1").mkdir()
sub1 = (test1 / "sub1").touch()

with context(scope="main"), monitor.transaction():
    admin = monitor.new_group("admin", 100)
    PE.root.set(admin, baz, NodeState("vma"))

    user = monitor.new_user("cithun")
    monitor.user_inherit(user, admin)

    assert PE(user).get(user, baz).most == NodeState("vma")
    assert not PE.root.get(user, qux).most.available

    PE.root.set(user, qux, NodeState(7))
    assert PE.root.get(user, qux).most.available

    PE.root.set(admin, baz, NodeState("v-a"))
    try:
        PE(user).set(user, qux, NodeState("vm-"))
    except PermissionError as e:
        print(e)  # Permission denied as /baz/ is not modifiable

    PE.root.set(user, Node("/command/*"), NodeState("vma"))
    assert PE(user).get(user, sub).most == NodeState("vma")



from __future__ import annotations

import fnmatch
import os
import posixpath
import re
import sys
from collections.abc import Sequence
from dataclasses import dataclass, field
from types import GenericAlias, MappingProxyType
from typing import List, Literal, Tuple, Union

_MAPPING = {"-": 0, "a": 1, "m": 2, "v": 4}


class NodeState:
    AVAILABLE = 1
    """
    若 state 所属的权限节点拥有子节点，则表示对应的权限拥有者默认情况下对该节点的子节点拥有使用权限
    
    若 state 所属的权限节点不拥有子节点，则表示对应的权限拥有者对该节点拥有使用权限，表示对节点对应的实际内容可用
    """

    MODIFY = 2
    """
    若 state 所属的权限节点拥有子节点，则表示对应的权限拥有者可以增加、删除、修改该节点的子节点（无论子节点的权限是怎样的）
    
    若 state 所属的权限节点不拥有子节点，则表示对应的权限拥有者可以修改该节点的内容
    """

    VISIT = 4
    """
    若 state 所属的权限节点拥有子节点，则表示对应的权限拥有者可以访问该节点的子节点
    
    若 state 所属的权限节点不拥有子节点，则表示对应的权限拥有者可以查看节点的状态和内容
    """

    def __init__(self, state: int | str):
        if isinstance(state, str):
            state = sum(_MAPPING[i] for i in state.lower() if i in _MAPPING)
        if state < 0 or state > 7:
            raise ValueError("state must be in range [0, 7]")
        self.state = state

    @property
    def available(self):
        return self.state & NodeState.AVAILABLE == NodeState.AVAILABLE

    @property
    def modify(self):
        return self.state & NodeState.MODIFY == NodeState.MODIFY

    @property
    def visit(self):
        return self.state & NodeState.VISIT == NodeState.VISIT

    def __repr__(self):
        state = ["-", "-", "-"]
        if self.available:
            state[2] = "a"
        if self.modify:
            state[1] = "m"
        if self.visit:
            state[0] = "v"
        return "".join(state)

    def __gt__(self, other):
        return self.state > other.state

    def __ge__(self, other):
        return self.state >= other.state

    def __lt__(self, other):
        return self.state < other.state

    def __le__(self, other):
        return self.state <= other.state

    def __eq__(self, other):
        return self.state == other.state

    def __hash__(self):
        return hash(self.state)

    def __bool__(self):
        return bool(self.state)

    def __int__(self):
        return self.state

    def __and__(self, other):
        return NodeState(self.state & other.state)

    def __or__(self, other):
        return NodeState(self.state | other.state)


CHILD_MAP: dict[str, set[str]] = {"/": set()}


@dataclass(repr=False, eq=True, unsafe_hash=True)
class File:
    type: Literal["file", "dir"]
    path: str
    content: dict[str, str] = field(compare=False, hash=False)

    def __post_init__(self):
        if self.type == "dir":
            CHILD_MAP[self.path] = set()

    @property
    def isdir(self):
        return self.type == "dir"

    @property
    def isfile(self):
        return self.type == "file"

    def __repr__(self):
        return f"{'DIR' if self.isdir else 'FILE'}({self.path!r})"


INDEX_MAP: dict[str, File] = {"/": File("dir", "/", {})}


class _Flavour(object):
    """A flavour implements a particular (platform-specific) set of node
    semantics."""

    def __init__(self):
        self.sep = "/"
        self.altsep = ""
        self.join = self.sep.join

    def parse_parts(self, parts: List[str]):
        parsed: List[str] = []
        sep = self.sep
        altsep = self.altsep
        root = ""
        it = reversed(parts)
        for part in it:
            if not part:
                continue
            if altsep:
                part = part.replace(altsep, sep)
            root, rel = self.splitroot(part)
            if sep in rel:
                for x in reversed(rel.split(sep)):
                    if x and x != ".":
                        parsed.append(sys.intern(x))
            else:
                if rel and rel != ".":
                    parsed.append(sys.intern(rel))
        if root:
            parsed.append(root)
        parsed.reverse()
        return root, parsed

    def join_parsed_parts(self, root, parts, root2, parts2):
        """
        Join the two nodes represented by the respective
        (root, parts) tuples.  Return a new (root, parts) tuple.
        """
        if not root2:
            # Second node is non-anchored (common case)
            return root, parts + parts2
        return root2, parts2

    def splitroot(self, part, sep="/"):
        if part and part[0] == sep:
            stripped_part = part.lstrip(sep)
            if len(part) - len(stripped_part) == 2:
                return sep * 2, stripped_part
            else:
                return sep, stripped_part
        else:
            return "", part

    def casefold(self, s: str):
        return s

    def casefold_parts(self, parts: List[str]):
        return parts

    def compile_pattern(self, pattern: str):
        return re.compile(fnmatch.translate(pattern)).fullmatch

    def resolve(self, node):
        sep = self.sep

        def _resolve(node, rest):
            if rest.startswith(sep):
                node = ""

            for name in rest.split(sep):
                if not name or name == ".":
                    # current dir
                    continue
                if name == "..":
                    # parent dir
                    node, _, _ = node.rpartition(sep)
                    continue
                if node.endswith(sep):
                    newnode = node + name
                else:
                    newnode = node + sep + name
                node = newnode

            return node

        base = "" if node.is_absolute() else _get_current()
        return _resolve(base, str(node)) or sep

    def is_reserved(self, parts):
        return False


_flavour = _Flavour()


class _NodeParents(Sequence):
    """This object provides sequence-like access to the logical ancestors
    of a node.  Don't try to construct it yourself."""

    __slots__ = ("_nodecls", "_root", "_parts")

    def __init__(self, node: "Node"):
        # We don't store the instance to avoid reference cycles
        self._nodecls = type(node)
        self._root = node._root
        self._parts = node._parts

    def __len__(self):
        if self._root:
            return len(self._parts) - 1
        else:
            return len(self._parts)

    def __getitem__(self, idx):
        if idx < 0 or idx >= len(self):
            raise IndexError(idx)
        return self._nodecls._from_parsed_parts(self._root, self._parts[: -idx - 1])

    def __repr__(self):
        return "<Node.parents>"


class Node:
    __slots__ = (
        "_root",
        "_parts",
        "_str",
        "_hash",
        "_pparts",
        "_cached_cparts",
    )

    _root: str
    _parts: list[str]
    _str: str
    _hash: int
    _pparts: tuple[str, ...]

    def __new__(cls, *args, **kwargs):
        self = cls._from_parts(args, init=False)
        self._init()
        return self

    def __reduce__(self):
        # Using the parts tuple helps share interned node parts
        # when pickling related nodes.
        return self.__class__, tuple(self._parts)

    @classmethod
    def _parse_args(cls, args: Sequence[Union[Node, str]]):
        # This is useful when you don't want to create an instance, just
        # canonicalize some constructor arguments.
        parts: List[str] = []
        for a in args:
            if isinstance(a, Node):
                parts += a._parts
            elif isinstance(a, str):
                # Force-cast str subclasses to str (issue #21127)
                parts.append(str(a))
            else:
                raise TypeError(f"argument should be a str, not {type(a)!r}")
        return _flavour.parse_parts(parts)

    @classmethod
    def _from_parts(cls, args: Sequence[Union[Node, str]], init: bool = True):
        self = object.__new__(cls)
        root, parts = self._parse_args(args)
        self._root = root
        self._parts = parts
        if init:
            self._init()
        return self

    @classmethod
    def _from_parsed_parts(cls, root: str, parts: List[str], init: bool = True):
        self = object.__new__(cls)
        self._root = root
        self._parts = parts
        if init:
            self._init()
        return self

    @classmethod
    def _format_parsed_parts(cls, root: str, parts: List[str]):
        if root:
            return root + _flavour.join(parts[1:])
        else:
            return _flavour.join(parts)

    def _make_child(self, args: Tuple[Union[Node, str], ...]):
        root, parts = self._parse_args(args)
        root, parts = _flavour.join_parsed_parts(self._root, self._parts, root, parts)
        return self._from_parsed_parts(root, parts)

    def dot(self):
        path = str(self)
        if path == "/":
            return ""
        if path.startswith("/"):
            path = path[1:]
        return path.replace(" ", "-").replace(".", "_").replace("/", ".")

    def __str__(self):
        """Return the string representation of the node, suitable for
        passing to system calls."""
        try:
            return self._str
        except AttributeError:
            self._str = self._format_parsed_parts(self._root, self._parts) or "."
            return self._str

    def __fspath__(self):
        return str(self)

    def __bytes__(self):
        """Return the bytes representation of the node.  This is only
        recommended to use under Unix."""
        return os.fsencode(self)

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, str(self))

    @property
    def _cparts(self):
        try:
            return self._cached_cparts
        except AttributeError:
            self._cached_cparts = _flavour.casefold_parts(self._parts)
            return self._cached_cparts

    def __eq__(self, other):
        if not isinstance(other, Node):
            raise NotImplementedError
        return self._cparts == other._cparts

    def __hash__(self):
        try:
            return self._hash
        except AttributeError:
            self._hash = hash(tuple(self._cparts))
            return self._hash

    def __lt__(self, other):
        if not isinstance(other, Node):
            raise NotImplementedError
        return self._cparts < other._cparts

    def __le__(self, other):
        if not isinstance(other, Node):
            raise NotImplementedError
        return self._cparts <= other._cparts

    def __gt__(self, other):
        if not isinstance(other, Node):
            raise NotImplementedError
        return self._cparts > other._cparts

    def __ge__(self, other):
        if not isinstance(other, Node):
            raise NotImplementedError
        return self._cparts >= other._cparts

    __class_getitem__ = classmethod(GenericAlias)  # noqa

    @property
    def root(self):
        """The root of the node, if any."""
        return self._root

    @property
    def name(self):
        """The final node component, if any."""
        parts = self._parts
        if len(parts) == (1 if self._root else 0):
            return ""
        return parts[-1]

    @property
    def suffix(self):
        """
        The final component's last suffix, if any.

        This includes the leading period. For example: '.txt'
        """
        name = self.name
        i = name.rfind(".")
        if 0 < i < len(name) - 1:
            return name[i:]
        else:
            return ""

    @property
    def suffixes(self):
        """
        A list of the final component's suffixes, if any.

        These include the leading periods. For example: ['.tar', '.gz']
        """
        name = self.name
        if name.endswith("."):
            return []
        name = name.lstrip(".")
        return ["." + suffix for suffix in name.split(".")[1:]]

    @property
    def stem(self):
        """The final node component, minus its last suffix."""
        name = self.name
        i = name.rfind(".")
        if 0 < i < len(name) - 1:
            return name[:i]
        else:
            return name

    def with_name(self, name: str):
        """Return a new node with the file name changed."""
        if not self.name:
            raise ValueError("%r has an empty name" % (self,))
        root, parts = _flavour.parse_parts([name])
        if not name or name[-1] in [_flavour.sep, _flavour.altsep] or root or len(parts) != 1:
            raise ValueError(f"Invalid name {name!r}")
        return self._from_parsed_parts(self._root, self._parts[:-1] + [name])

    def with_stem(self, stem: str):
        """Return a new node with the stem changed."""
        return self.with_name(stem + self.suffix)

    def with_suffix(self, suffix: str):
        """Return a new node with the file suffix changed.  If the node
        has no suffix, add given suffix.  If the given suffix is an empty
        string, remove the suffix from the node.
        """
        f = _flavour
        if f.sep in suffix or f.altsep and f.altsep in suffix:
            raise ValueError(f"Invalid suffix {suffix!r}")
        if suffix and not suffix.startswith(".") or suffix == ".":
            raise ValueError(f"Invalid suffix {suffix!r}")
        name = self.name
        if not name:
            raise ValueError(f"{self!r} has an empty name")
        old_suffix = self.suffix
        if not old_suffix:
            name += suffix
        else:
            name = name[: -len(old_suffix)] + suffix
        return self._from_parsed_parts(self._root, self._parts[:-1] + [name])

    def relative_to(self, *other):
        """Return the relative node to another node identified by the passed
        arguments.  If the operation is not possible (because this is not
        a subnode of the other node), raise ValueError.
        """

        if not other:
            raise TypeError("need at least one argument")
        parts = self._parts
        root = self._root
        if root:
            abs_parts = [root] + parts[1:]
        else:
            abs_parts = parts
        to_root, to_parts = self._parse_args(other)
        if to_root:
            to_abs_parts = [to_root] + to_parts[1:]
        else:
            to_abs_parts = to_parts
        n = len(to_abs_parts)
        cf = _flavour.casefold_parts
        if root if n == 0 else cf(abs_parts[:n]) != cf(to_abs_parts):
            formatted = self._format_parsed_parts(to_root, to_parts)
            raise ValueError(
                "{!r} is not in the subnode of {!r}"
                " OR one node is relative and the other is absolute.".format(str(self), str(formatted))
            )
        return self._from_parsed_parts(root if n == 1 else "", abs_parts[n:])

    def is_relative_to(self, *other):
        """Return True if the node is relative to another node or False."""
        try:
            self.relative_to(*other)
            return True
        except ValueError:
            return False

    @property
    def parts(self):
        """An object providing sequence-like access to the
        components in the filesystem node."""
        # We cache the tuple to avoid building a new one each time .parts
        # is accessed.  XXX is this necessary?
        try:
            return self._pparts
        except AttributeError:
            self._pparts = tuple(self._parts)
            return self._pparts

    def joinpath(self, *args: Union[Node, str]) -> Node:
        """Combine this node with one or several arguments, and return a
        new node representing either a subnode (if all arguments are relative
        nodes) or a totally different node (if one of the arguments is
        anchored).
        """
        return self._make_child(args)

    def __truediv__(self, key: str):
        try:
            return self._make_child((key,))
        except TypeError:
            raise NotImplementedError

    def __rtruediv__(self, key: str):
        try:
            return self._from_parts([key] + self._parts)
        except TypeError:
            raise NotImplementedError

    @property
    def parent(self):
        """The logical parent of the node."""
        root = self._root
        parts = self._parts
        if len(parts) == 1 and root:
            return self
        return self._from_parsed_parts(root, parts[:-1])

    @property
    def parents(self):
        """A sequence of this node's logical parents."""
        return _NodeParents(self)

    def is_absolute(self):
        """True if the node is absolute (has both a root and, if applicable,
        a drive)."""
        return bool(self._root)

    def is_reserved(self):
        """Return True if the node contains one of the special names reserved
        by the system, if any."""
        return _flavour.is_reserved(self._parts)

    def match(self, node_pattern):
        """
        Return True if this node matches the given pattern.
        """
        cf = _flavour.casefold
        node_pattern = cf(node_pattern)
        root, pat_parts = _flavour.parse_parts([node_pattern])
        if not pat_parts:
            raise ValueError("empty pattern")
        if root and root != cf(self._root):
            return False
        parts = self._cparts
        if root:
            if len(pat_parts) != len(parts):
                return False
            pat_parts = pat_parts[1:]
        elif len(pat_parts) > len(parts):
            return False
        for part, pat in zip(reversed(parts), reversed(pat_parts)):
            if not fnmatch.fnmatchcase(part, pat):
                return False
        return True

    def _init(self): ...

    def _make_child_relnode(self, part: str):
        # This is an optimization used for dir walking.  `part` must be
        # a single part relative to this node.
        parts = self._parts + [part]
        return self._from_parsed_parts(self._root, parts)

    # Public API

    def absolute(self):
        """Return an absolute version of this node.  This function works
        even if the node doesn't point to anything.

        No normalization is done, i.e. all '.' and '..' will be kept along.
        Use resolve() to get the canonical node to a file.
        """
        if self.is_absolute():
            return self
        obj = self._from_parts([_get_current()] + self._parts, init=False)
        obj._init()
        return obj

    def resolve(self):
        """
        Make the node absolute, resolving all symlinks on the way and also
        normalizing it (for example turning slashes into backslashes under
        Windows).
        """
        s = _flavour.resolve(self)
        normed = posixpath.normpath(s)
        obj = self._from_parts((normed,), init=False)
        obj._init()
        return obj

    def content(self):
        return INDEX_MAP[str(self.absolute())].content

    def samefile(self, other_node):
        """Return whether other_node is the same or not as this file
        (as returned by os.node.samefile()).
        """
        st = self.content()
        try:
            other_st = other_node.content()
        except KeyError:
            return False
        return st == other_st

    def read(self):
        return MappingProxyType(self.content())

    def write(self, content: dict[str, str], create_if_not_exists=False):
        if not self.exists():
            if create_if_not_exists:
                self.touch(content)
            else:
                raise FileNotFoundError(f"No such file or directory: '{self}'")
        if self.is_dir():
            raise IsADirectoryError(f"Is a directory: '{self}'")
        self.content().update(content)

    def touch(self, content: dict[str, str] | None = None, exist_ok=True):
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        node = self.absolute()
        if str(node) in INDEX_MAP:
            if not exist_ok:
                raise FileExistsError(f"File exists: '{node}'")
            return self
        if not node.parent.exists():
            raise FileNotFoundError(f"No such file or directory: '{node.parent}'")
        INDEX_MAP[str(node)] = File("file", str(node), content or {})
        CHILD_MAP[str(node.parent)].add(node.name)
        return self

    def mkdir(self, parents=False, exist_ok=False):
        """
        Create a new directory at this given node.
        """
        if self.exists():
            if not exist_ok or not self.is_dir():
                raise FileExistsError(f"File exists: '{self}'")
            return self
        if not self.parent.exists():
            if not parents:
                raise FileNotFoundError(f"No such file or directory: '{self.parent}'")
            self.parent.mkdir(parents=True, exist_ok=True)
        INDEX_MAP[str(self)] = File("dir", str(self), {})
        CHILD_MAP[str(self.parent)].add(self.name)
        return self

    def iterdir(self):
        """Iterate over the files in this directory.  Does not yield any
        result for the special nodes '.' and '..'.
        """
        if not self.is_dir():
            raise NotADirectoryError(f"Not a directory: '{self}'")
        for name in CHILD_MAP[str(self)]:
            if name in {".", ".."}:
                # Yielding a node object for these makes little sense
                continue
            yield self._make_child_relnode(name)

    def exists(self):
        """
        Whether this node exists.
        """
        try:
            self.content()
        except KeyError:
            return False
        return True

    def is_dir(self):
        """
        Whether this node is a directory.
        """
        if not self.exists():
            raise FileNotFoundError(f"No such file or directory: '{self}'")
        return INDEX_MAP[str(self.absolute())].isdir

    def is_file(self):
        """
        Whether this node is a regular file (also True for symlinks pointing
        to regular files).
        """
        if not self.exists():
            raise FileNotFoundError(f"No such file or directory: '{self}'")
        return not INDEX_MAP[str(self.absolute())].isdir


ROOT = Node("/")
_CURRENT: "Node" = ROOT


def _get_current():
    return str(_CURRENT)


def _set_current(node: "Node"):
    global _CURRENT
    _CURRENT = node


def mkdir(path: str, base: Node = ROOT, parents: bool = False, exist_ok: bool = False) -> Node:
    return (base / path).mkdir(parents, exist_ok)


def touch(path: str, base: Node = ROOT, content: dict[str, str] | None = None, exist_ok: bool = True) -> Node:
    return (base / path).touch(content, exist_ok)



from abc import ABC, abstractmethod
from typing import Iterable

from .owner import Group, User


class SyncMonitor(ABC):
    @abstractmethod
    def new_group(self, name: str, priority: int) -> Group:
        pass

    @abstractmethod
    def new_user(self, name: str) -> User:
        pass

    @abstractmethod
    def load(self):
        pass

    @abstractmethod
    def save(self):
        pass

    @abstractmethod
    def group_inherit(self, target: Group, *groups: Group):
        pass

    @abstractmethod
    def user_inherit(self, target: User, *groups: Group):
        pass

    @abstractmethod
    def user_leave(self, target: User, group: Group):
        pass

    @abstractmethod
    def all_users(self) -> Iterable[User]:
        pass

    @abstractmethod
    def all_groups(self) -> Iterable[Group]:
        pass


class AsyncMonitor(ABC):
    @abstractmethod
    async def new_group(self, name: str, priority: int) -> Group:
        pass

    @abstractmethod
    async def new_user(self, name: str) -> User:
        pass

    @abstractmethod
    async def load(self):
        pass

    @abstractmethod
    async def save(self):
        pass

    @abstractmethod
    async def group_inherit(self, target: Group, *groups: Group):
        pass

    @abstractmethod
    async def user_inherit(self, target: User, *groups: Group):
        pass

    @abstractmethod
    async def user_leave(self, target: User, group: Group):
        pass

    @abstractmethod
    async def all_users(self) -> Iterable[User]:
        pass

    @abstractmethod
    async def all_groups(self) -> Iterable[Group]:
        pass



import json
from contextlib import contextmanager
from pathlib import Path
from typing import Iterable
from weakref import WeakValueDictionary

from arclet.cithun import Group, User
from arclet.cithun.monitor import SyncMonitor

from .owner import DefaultGroup, DefaultUser


class DefaultMonitor(SyncMonitor):
    def __init__(self, file: Path):
        if not file.suffix.startswith(".json"):
            raise ValueError(file)
        self.file = file
        self.USER_TABLE = WeakValueDictionary()
        self.GROUP_TABLE = WeakValueDictionary()

    def new_group(self, name: str, priority: int):
        if name in self.GROUP_TABLE:
            raise ValueError(f"Group {name} already exists")
        group = DefaultGroup(name, priority)
        self.GROUP_TABLE[name] = group
        return group

    def new_user(self, name: str):
        if name in self.USER_TABLE:
            raise ValueError(f"User {name} already exists")
        user = DefaultUser(name)
        self.USER_TABLE[name] = user
        return user

    def load(self):
        if self.file.exists():
            with self.file.open("r", encoding="utf-8") as f:
                data = json.load(f)
            users = {name: DefaultUser.parse(raw) for name, raw in data["users"].items()}
            groups = {name: DefaultGroup.parse(raw) for name, raw in data["groups"].items()}
            self.USER_TABLE.update(users)
            self.GROUP_TABLE.update(groups)
            for group in groups.values():
                group.inherits = [self.GROUP_TABLE[gp.name] for gp in group.inherits]
            for user in users.values():
                user.inherits = [self.GROUP_TABLE[gp.name] for gp in user.inherits]
            del users, groups
        else:
            with self.file.open("w+", encoding="utf-8") as f:
                json.dump({}, f, ensure_ascii=False)

    def save(self):
        data = {
            "users": {user.name: user.dump() for user in self.USER_TABLE.values()},
            "groups": {group.name: group.dump() for group in self.GROUP_TABLE.values()},
        }
        with self.file.open("w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=2)

    def group_inherit(self, target: Group, *groups: Group):
        for group in groups:
            if group not in target.inherits:
                target.inherits.append(group)

    def user_inherit(self, target: User, *groups: Group):
        for group in groups:
            if group not in target.inherits:
                target.inherits.append(group)

    def user_leave(self, target: User, group: Group):
        if group in target.inherits:
            target.inherits.remove(group)

    @contextmanager
    def transaction(self):
        yield
        self.save()

    def all_users(self) -> Iterable[User]:
        return self.USER_TABLE.values()

    def all_groups(self) -> Iterable[Group]:
        return self.GROUP_TABLE.values()



from __future__ import annotations

from dataclasses import dataclass, field

from arclet.cithun.ctx import Context
from arclet.cithun.node import Node, NodeState


@dataclass(eq=True, unsafe_hash=True)
class DefaultGroup:
    name: str
    priority: int
    nodes: dict[Node, dict[Context, NodeState]] = field(default_factory=dict, compare=False, hash=False)
    inherits: list = field(default_factory=list, compare=False, hash=False)

    def dump(self):
        return {
            "name": self.name,
            "priority": self.priority,
            "nodes": {
                str(node): {str(ctx): state.state for ctx, state in data.items()} for node, data in self.nodes.items()
            },
            "inherits": [gp.name for gp in self.inherits],
        }

    @classmethod
    def parse(cls, raw: dict):
        obj = cls(raw["name"], raw["priority"])
        obj.nodes = {
            Node(node): {Context.from_string(ctx): NodeState(state) for ctx, state in data.items()}
            for node, data in raw["nodes"].items()
        }
        obj.inherits = [DefaultGroup(name, 0) for name in raw["inherits"]]
        return obj

    def __str__(self):
        return f"Group({self.name})"


@dataclass(eq=True, unsafe_hash=True)
class DefaultUser:
    name: str
    nodes: dict[Node, dict[Context, NodeState]] = field(default_factory=dict, compare=False, hash=False)
    inherits: list = field(default_factory=list, compare=False, hash=False)

    def dump(self):
        return {
            "name": self.name,
            "nodes": {
                str(node): {str(ctx): state.state for ctx, state in data.items()} for node, data in self.nodes.items()
            },
            "inherits": [user.name for user in self.inherits],
        }

    @classmethod
    def parse(cls, raw: dict):
        obj = cls(raw["name"])
        obj.nodes = {
            Node(node): {Context.from_string(ctx): NodeState(state) for ctx, state in data.items()}
            for node, data in raw["nodes"].items()
        }
        obj.inherits = [DefaultUser(name) for name in raw["inherits"]]
        return obj

    def __str__(self):
        return f"User({self.name})"
